(*  Abrimos los modulos que contienen funciones de apoyo a las funciones de este fichero *)
#open "lexing";;
#open "fichia";;
#open "scanord";;
#open "parord";;

(*  Definicion del tipo respuesta_back *)

type respuesta_back = Fallo | S of ((string * term) list);;


(* Funcion auxiliar de leer_entrada que lee la cadena de entrada *)

let lee_expresion ()=
	Entrada token_ord (create_lexer_string (read_line()^"\n"));;

(* Funciones encargadas de imprimir de forma agradable los terminos ,los atomos , las substituciones y las expresiones *) 

let rec pretty_term = fun
	  (Var s) -> print_string s
	| (Con s) -> print_string s;;
 
let pretty_ato = function
	  At(s,[])   -> print_string s
	| At(s,t::l) -> print_string s;
			print_string "(";
			pretty_term t;
			do_list (fun x -> print_string ",";
				 pretty_term x) l;
			print_string ")";;


let pretty_subst = 
	let pretty_par = function
	(x,y) -> print_string "[ ";
		 pretty_term y;
		 print_string " <- ";
		 print_string x;
		 print_string " ]" in
	function 
	  Fallo   -> ()
	| S([])   -> print_string "{ ";
			    print_string " }";
	| S(h::l) -> print_string "{ ";
		     pretty_par h;
		     do_list (fun x -> print_string ",";pretty_par x) l;
		     print_string " }";
		     print_newline ();;


let pretty_lista_subst = function
	  []          -> ()
	| (h::l) as t -> do_list (fun x -> pretty_subst x) t;;

let rec pretty_exp = function
	  Exp (x,[])   -> pretty_ato x; print_newline ()
	| Exp (x,h::l) -> pretty_ato h;
			  do_list (fun x -> print_string " ^ ";pretty_ato x) l;
			  print_string " -> ";
			  pretty_ato x;
			  print_newline ();;

let pretty_lista_exp = function
	  []          -> ()
	| (h::l) as t -> do_list (fun x -> pretty_exp x) t;;



(* Funcion que realiza una substitucion sobre un termino *)

let sustituye_termino sust=function Con x -> Con x
				   |Var(s) as t -> try assoc s sust with Not_found -> t;;

(* Funcion que realiza la substitucion sobre un atomo *)

let subst_rec subst=function At(nom,terms)->At(nom,map (sustituye_termino subst) terms);;

(* Funcion que realiza una substitucion sobre una expresion *)

let substitute subst= function Exp(atomica,anteceds) -> Exp((subst_rec subst) atomica, map (subst_rec subst) anteceds) ;;

(* La siguiente funcion realiza la substitucion de un termino en una lista de expresiones *)
let subslista s l=map (substitute s) l;;

(* Funcion que realiza la substitucion de un termino sobre una lista de expresiones *) 
let subat s l=map (subst_rec s) l;;

(* La siguiente funcion realiza la composicion de dos substituciones *)
let compsubst subst1 subst2 = 
  encadena (map (fun (v,t) -> (v, sustituye_termino subst1 t)) subst2) subst1 
where rec encadena s1 t= if t=[] then s1 else
			    if mem (hd t) s1 then encadena s1 (tl t)
					     else encadena ((hd t)::s1) (tl t) 
;;

(* Funcion auxiliar que realiza la unificacion de dos listas de terminos y devuelve una lista con las substituciones o un error*)
let rec unify_terms = fun [] [] ac -> ac
			|(ta::t1) (tb::t2) ac ->(if (ta=tb) then unify_terms t1 t2 ac
						else match (ta,tb) with
								Con _,Con _->failwith "unify"
								|Con v,Var s->(try 
										let T=(assoc s ac) in 
										 if T=Con v then unify_terms t1 t2 ac
										 else failwith "unify"
										with Not_Found -> 
										 unify_terms t1 t2 (compsubst [(s,Con v)] ac))
								|Var s,Con v->(try 
										let T=(assoc s ac) in 
										 if T=Con v then unify_terms t1 t2 ac
										 else failwith "unify"
										with Not_Found -> 
										 unify_terms t1 t2 (compsubst [(s,Con v)] ac))
								|Var s,Var v->unify_terms t1 t2 (compsubst ac [(s,Var v)]))
			|_ _ _->failwith "unify"
			;;


(* Funcion que comprueba la cabeza de los atomos y el numero de argumentos de los mismos para confirmar su posible unificacion *)
let unify= fun (Exp(At(s,t1),[])) (Exp(At(t,t2),[])) -> 
			if ((s=t) && ((list_length t1) = (list_length t2))) then unify_terms t1 t2 []
			else failwith "unify"
		|_ _ -> failwith "unify"
;; 

(* La siguiente funcion cuando se le pasa una lista de terminos devuelve otra lista que se corresponde con las variables distintas encontradas en los terminos *)
let rec vars_terms= fun [] -> []
			|((Var x)::l) ->union [x] (vars_terms l)
			|(_::l) -> vars_terms l
;;

(* La funcion variables devuelve la lista de variables distintas que hay en una expresion, esta funcion utiliza la funcion auxiliar vars_terms *)
let variables = 
	fun (Exp(At(_,ls),ant)) ->
		union (var_lista_atomicas ant) (vars_terms ls) 
			where rec var_lista_atomicas lista = if lista =[] then []
							     else 
								match (hd lista) with 
								At (_,lt) -> union (vars_terms lt) (var_lista_atomicas (tl lista)) 
						
						 
;;

(* La siguiente funcion realiza la interseccion de dos listas, aunque esta funcion realmente ya la incorpora el camllight *)

let rec intersection l1 l2 = if l1=[] then []
		 	     else if (mem (hd l1) l2) then union [(hd l1)] (intersection (tl l1) l2)
				  else intersection (tl l1) l2
;;

(* La funcion unificar es una funcion ampliada de la funcion unify, vista anteriormente, que comprueba si exp1 y exp2 tienen variables comunes y si esto es cierto entonces renombra las variables de la exp2 *)

let unificar exp1 exp2 = 
	let inter=intersection (variables exp1) (variables exp2) in

	if inter=[] then
		(unify exp1 exp2 , [])
	else
	    let rec cambia2 comunes exp2cambiada = 
	      if comunes = [] then exp2cambiada
	      else 
	       (cambia2 (tl comunes) [(hd comunes),Var ((hd comunes)^"X")])

	    in 
		let ren=(cambia2 inter []) in ((unify exp1 e2,ren) where e2=substitute ren exp2 )
				     ;;


(* Funcion auxiliar de unify_terms que realiza la unificacion de dos terminos *)
let rec unify_t = fun [] [] ac -> ac
			|(ta::t1) (tb::t2) ac ->(if (ta=tb) then unify_t t1 t2 ac
						else match (ta,tb) with
								Con _,Con _->failwith "unify"
								|Con v,Var s->failwith "unify"                                          
								|Var s,Con v->failwith "unify"  
								|Var s,Var v->
									(try 
									    let T=(assoc s ac) in 
									     if T=Var v then unify_t t1 t2 ac
									     else failwith "unify"
									with Not_found -> unify_t t1 t2 ([(s,Var v)]@ac)
									     | _ -> failwith "unify" ))
			|_ _ _ ->failwith "unify"
			;;


(* Funcion que comprueba si todos los elementos de una lista son variables, devolviendo true en este caso *)
let rec todas_variables =function [] -> true
				 |(_,Var _)::l->todas_variables l
				 |(_,Con _)::l-> false
;;

(* Definicion de la base de conocimiento y de la lista auxiliar answers *)
let base=ref [];;
let answers=ref [];;

(* Funcion que comprueba si una expresion contiene solamente un atomo o por el contrario si es una lista de antecedentes y un consecuente *)

let rec cumple_condicion atomica = 
	function (Exp(_,[]))->[(false,[],[],[])]
		|Exp(q,l)->
			let rec auxiliar consecuente ac =
				function (x::y) -> 
			(try 
			 ((true,subat d (ac@y),sigma,[subst_rec d consecuente]) :: (auxiliar consecuente (x::ac) y))
			 where (sigma,d)=(unificar (Exp(x,[])) atomica)
			with _ ->(auxiliar consecuente (x::ac) y))
			|[] -> [(false,[],[],[])]
			in auxiliar q [] l
;;


let si = fun _ -> true;;

let invierte (a,b) = (b,a) ;;


(* Funcion auxiliar de la funcion esta *)
let rec comprueba lista ac = if lista=[] then ac
			     else 
				try
				 let x=assoc (snd(hd lista)) (map invierte (tl lista)) in failwith "unify"
				with Not_found -> comprueba (tl lista) ac ;;


(* La siguiente expresion comprueba si una expresion se encuentra en una lista *)

let rec esta e lista = 
		       if lista=[] then false

		       else 

			(let rec compr exp1 exp2 ac = 
		
			   match (exp1,exp2)
				with 
				   ( (Exp(At(n,lt1),[])),(Exp(At(m,lt2),[]))) ->
					(
					if m=n then     
						try
						  si (comprueba (unify_t lt1 lt2 ac) (unify_t lt1 lt2 ac))
						with _ -> false
					else false
					)
 
				   |((Exp(At(n,lt1),a::b)),(Exp(At(m,lt2),c::d))) ->
				     (if m=n then 
						try
						  compr (substitute s (Exp(a,b))) (substitute s (Exp(c,d))) s 
						  where s=comprueba (unify_t lt1 lt2 ac) (unify_t lt1 lt2 ac)
						with _ -> false
				     else false)
				   |(_,_) -> false
				   
							

			in if (compr e (hd lista) []) then true else esta e (tl lista));;



(* La siguiente funcion realiza el algoritmo forward chaining *)

let rec encadenamiento_forward atomica = 

let rec find_and_infer (KB,premises,conclusion,sustitucion)  = 

	(if premises=[] then 
		begin
			let x =  (substitute sustitucion (Exp(conclusion,[]))) in
			  if (esta x KB) then []
			  else
			  (
			  encadenamiento_forward x
			  )
		end
	 else
	let rec recorre lista  = 
		if lista=[] then []
		else
		  try
		   union (find_and_infer(KB,(tl premises),conclusion,compsubst sigma2 sustitucion)) (recorre (tl lista))
		   where (sigma2,l)=(unificar (substitute sustitucion (Exp((hd premises),[]))) (hd lista)) 
		  with _ -> recorre (tl lista)
	in recorre KB)

in      (       
	if esta atomica (!base) then []
		else 
		  let rec enc_aux lista_total=function [] -> []
						  | (x::y)->
						    let l=(cumple_condicion atomica x) in
							let rec recorre = fun [] -> []
									    |((a,b,c,d)::r) ->
										(if a then 
				(let k=(union ((find_and_infer ((!base),b,hd d,c))@(enc_aux lista_total y)) (recorre r)) in
					(base:=union k (!base);k))
									else (enc_aux lista_total y))
							in recorre l


		  in (base:=(atomica::(!base));(atomica::(enc_aux (!base) (!base)) )));;


let at_to_exp x=Exp (x,[]);;

(* La siguiente funcion elimina un elemento de una lista *) 
let rec elimina lista elem = if lista=[] then []
			     else if (hd lista) = elem then (tl lista)
				  else (hd lista)::elimina (tl lista) elem
;;


(* La funcion en_answers comprueba si existe la sustitucion subst en la lista answers *)
let rec en_answers sust = function [] -> false
				|(a::b) -> if igual a sust then true
					   else en_answers sust b 
			  where rec igual = fun      (S[]) ([]) -> true
						     |(S(a::b)) ((c::d) as t) -> if mem a t then igual (S b) (elimina t a) else false
						     | _ _ -> false
;;


(* La proxima funcion implementa el algoritmo de backward chaining *)
let rec encadenamiento_back lista objetivo = back_chain_list lista [objetivo] [] 
				
where rec back_chain_list kabe qlist sigma = 
	answers:=[];
	if qlist=[] then ([S sigma])
	else let q=hd qlist in
	      (
	       let rec recorre lista = if lista=[] then ()
				       else
					try
					 if (en_answers ((compsubst sigma2 sigma)) (!answers)) then recorre (tl lista)
					 else
					  answers:=((S(compsubst sigma2 sigma))::(!answers));recorre (tl lista)
					 where sigma2=(unify (hd lista) q) 
					with _ -> recorre (tl lista)
					in recorre kabe) ;
					let rec recorre2 lista = 
					  if lista=[] then ()
					  else
					    match (hd lista) with Exp(x,[]) -> 
						recorre2 (tl lista)
						|Exp(x,(y::z)) ->   try
									   answers:=(union 
						(back_chain_list kabe (subslista sigma2 (map at_to_exp (y::z))) (compsubst sigma2 sigma))
											  (!answers));recorre2 (tl lista)
													
										where sigma2=unify q (at_to_exp x)
									  with _ -> recorre2 (tl lista)
							in recorre2 kabe;
							let rec uniones kb resto ans = 
								if ans=[] then [] 
								else 
								   match (hd ans) with
									(S x) -> union (back_chain_list kb (subslista (x) resto) x)
												(uniones kb resto (tl ans))
									|_ -> (uniones kb resto (tl ans)) 
									
							in (uniones kabe (tl qlist) !answers);;


(* La funcion save_aux es la funcion que implementa el comando save del sistema *)
let save_aux sentencia = match sentencia 
			 with (Exp(x,[])) ->
					let x=(encadenamiento_forward sentencia) in
						( print_string "Number of new clausules in K.B.:";
						  print_int (list_length x);
						  print_newline() 
						)
				|(Exp(x,y) as d)->if esta d (!base) then (print_string "It is already in the K.B.") 
					         else base:=sentencia::!base;print_string "A new implication entered the K.B."
								;print_newline();;


(* La funcion drop_aux implementa el comando drop que se utiliza en el sistema *)
let rec drop_aux lista sentencia = if lista=[] then 
					begin
						print_string "It wasn't in the K.B.";
						print_newline();
						[]
					end
				   else if (hd lista)=sentencia then (tl lista) 
					else (hd lista)::(drop_aux (tl lista) sentencia) ;;



(* Funcion que se encarga de leer la cadena de entrada y mostrar un mensaje si se produce un error *)
let rec leer_entrada () = try 
				lee_expresion ()
			  with _ -> (print_string "Input error\n>>";leer_entrada ())
;;



(* La siguiente funcion implementa el comando in utilizado en el sistema. Solo se puede invocar a la funcion in_aux usando una sentencia atomica como argumento *)

let rec in_aux e n = 
	fun [] -> (print_string "There were ";print_int n;print_string " unifications";
		   print_newline();flush std_out)
	  |(x::yy) -> match x with Exp (k,y::z) ->in_aux e n yy
		  		  |Exp (k,[])-> try 
						let (k,l)=unificar x e in 
						  if l=[] then 
					           (pretty_exp (substitute k e);in_aux e (n+1) yy)
						  else 
						   let e_changed=substitute l e in
						       (pretty_exp (substitute k e_changed);
							in_aux e (n+1) yy)
						with _ ->in_aux e n yy
;;


(* La funcion menu se encarga de procesar el comando recogido por teclado usando leer_entrada, y llamar a la funcion asociada *)
let rec menu () = 
		print_newline();
		print_string ">>";
	
		(match (leer_entrada ())
			    with
				ORD (List,[]) -> pretty_lista_exp !base ; menu ()
				|ORD (Save,[e]) -> save_aux e ; menu ()
				|ORD (Drop,[e]) -> base:=drop_aux !base e ; menu ()
				|ORD (Empty,[]) -> base:= [] ; menu ()
				|ORD (In,[e]) -> in_aux e 0 (!base) ; menu ()
				|ORD (Salir,[]) -> (print_string "Goodbye!";flush std_out;print_newline())
				|ORD (Proof,[e]) -> pretty_lista_subst (encadenamiento_back (!base) e); menu()
				|_ -> menu()
		)
	 
	 

in 

print_string "Hello!, this is the program for knowledge representation, ";
print_newline();
print_string " based in logic of first order";
print_newline();
print_string "You can save atomic sentences or implications (Horn clausules) ";
print_newline();
print_string "Introduce first the implication, and later the facts";
print_newline();
print_string "So on, the facts can produce the aparition of new facts ";
menu ();;











